# pylint: disable=unused-argument, arguments-differ, unused-variable
"""Fast-SCNN, implemented in Gluon. Code adapted from lxtGH/Fast_Seg"""
__all__ = ['FastSCNN', 'get_fastscnn', 'get_fastscnn_citys']

from mxnet.context import cpu
from mxnet.gluon.block import HybridBlock
from mxnet.gluon import nn


class FastSCNN(HybridBlock):
    r"""Fast-SCNN: Fast Semantic Segmentation Network

    Parameters
    ----------
    nclass : int
        Number of categories for the training dataset.
    norm_layer : object
        Normalization layer used in backbone network (default: :class:`mxnet.gluon.nn.BatchNorm`).
    aux : bool
        Auxiliary loss.


    Reference:

        Rudra P K Poudel, et al. https://bmvc2019.org/wp-content/uploads/papers/0959-paper.pdf
        "Fast-SCNN: Fast Semantic Segmentation Network." *BMVC*, 2019

    """
    def __init__(self, nclass, aux=True, ctx=cpu(), pretrained_base=False,
                 height=None, width=None, base_size=2048, crop_size=1024, **kwargs):
        super(FastSCNN, self).__init__()

        height = height if height is not None else crop_size
        width = width if width is not None else crop_size
        self._up_kwargs = {'height': height, 'width': width}
        self.base_size = base_size
        self.crop_size = crop_size

        self.aux = aux
        with self.name_scope():
            self.learning_to_downsample = LearningToDownsample(32, 48, 64, **kwargs)
            self.learning_to_downsample.initialize(ctx=ctx)
            self.global_feature_extractor = GlobalFeatureExtractor(
                64, [64, 96, 128], 128, 6, [3, 3, 3], height=height//32, width=width//32, **kwargs)
            self.global_feature_extractor.initialize(ctx=ctx)
            self.feature_fusion = FeatureFusionModule(64, 128, 128,
                                                      height=height//8, width=width//8, **kwargs)
            self.feature_fusion.initialize(ctx=ctx)
            self.classifier = Classifer(128, nclass, **kwargs)
            self.classifier.initialize(ctx=ctx)

            if self.aux:
                self.auxlayer = _auxHead(in_channels=64, channels=64, nclass=nclass, **kwargs)
                self.auxlayer.initialize(ctx=ctx)
                self.auxlayer.collect_params().setattr('lr_mult', 10)

    def hybrid_forward(self, F, x):
        """hybrid forward for Fast SCNN"""
        higher_res_features = self.learning_to_downsample(x)
        x = self.global_feature_extractor(higher_res_features)
        x = self.feature_fusion(higher_res_features, x)
        x = self.classifier(x)
        x = F.contrib.BilinearResize2D(x, **self._up_kwargs)

        outputs = []
        outputs.append(x)
        if self.aux:
            auxout = self.auxlayer(higher_res_features)
            auxout = F.contrib.BilinearResize2D(auxout, **self._up_kwargs)
            outputs.append(auxout)
        return tuple(outputs)

    def demo(self, x):
        """fastscnn demo"""
        h, w = x.shape[2:]
        self._up_kwargs['height'] = h
        self._up_kwargs['width'] = w
        self.global_feature_extractor.ppm._up_kwargs = {'height': h // 32, 'width': w // 32}
        self.feature_fusion._up_kwargs = {'height': h // 8, 'width': w // 8}

        higher_res_features = self.learning_to_downsample(x)
        x = self.global_feature_extractor(higher_res_features)
        x = self.feature_fusion(higher_res_features, x)
        x = self.classifier(x)
        import mxnet.ndarray as F
        x = F.contrib.BilinearResize2D(x, **self._up_kwargs)
        return x

    def predict(self, x):
        """fastscnn predict"""
        return self.demo(x)

    def evaluate(self, x):
        """evaluating network with inputs and targets"""
        return self.forward(x)[0]


class FeatureFusionModule(HybridBlock):
    """FastSCNN feature fusion module"""
    def __init__(self, highter_in_channels, lower_in_channels, out_channels, height, width,
                 scale_factor=4, norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(FeatureFusionModule, self).__init__()
        self.scale_factor = scale_factor
        self._up_kwargs = {'height': height, 'width': width}

        with self.name_scope():
            self.dwconv = _DWConv(lower_in_channels, out_channels,
                                  norm_layer=norm_layer, norm_kwargs=norm_kwargs)
            self.conv_lower_res = nn.Conv2D(in_channels=out_channels, channels=out_channels,
                                            kernel_size=1)
            self.conv_higher_res = nn.Conv2D(in_channels=highter_in_channels,
                                             channels=out_channels, kernel_size=1)
            self.bn = norm_layer(in_channels=out_channels)
            self.relu = nn.Activation('relu')

    def hybrid_forward(self, F, higher_res_feature, lower_res_feature):
        lower_res_feature = F.contrib.BilinearResize2D(lower_res_feature, **self._up_kwargs)

        lower_res_feature = self.dwconv(lower_res_feature)
        lower_res_feature = self.bn(self.conv_lower_res(lower_res_feature))
        higher_res_feature = self.bn(self.conv_higher_res(higher_res_feature))
        out = higher_res_feature + lower_res_feature
        return self.relu(out)


class Classifer(HybridBlock):
    """FastSCNN classifier"""
    def __init__(self, dw_channels, num_classes, stride=1,
                 norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(Classifer, self).__init__()
        with self.name_scope():
            self.dsconv1 = _DSConv(dw_channels, dw_channels, stride,
                                   norm_layer=norm_layer, norm_kwargs=norm_kwargs)
            self.dsconv2 = _DSConv(dw_channels, dw_channels, stride,
                                   norm_layer=norm_layer, norm_kwargs=norm_kwargs)
            self.dp = nn.Dropout(0.1)
            self.conv = nn.Conv2D(in_channels=dw_channels, channels=num_classes, kernel_size=1)

    def hybrid_forward(self, F, x):
        x = self.dsconv1(x)
        x = self.dsconv2(x)
        x = self.dp(x)
        x = self.conv(x)
        return x


class _auxHead(HybridBlock):
    # pylint: disable=redefined-outer-name
    def __init__(self, in_channels=64, channels=64, nclass=19,
                 norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(_auxHead, self).__init__()
        with self.name_scope():
            self.block = nn.HybridSequential()
            with self.block.name_scope():
                self.block.add(nn.Conv2D(in_channels=in_channels, channels=channels,
                                         kernel_size=3, padding=1, use_bias=False))
                self.block.add(norm_layer(in_channels=channels,
                                          **({} if norm_kwargs is None else norm_kwargs)))
                self.block.add(nn.Activation('relu'))
                self.block.add(nn.Dropout(0.1))
                self.block.add(nn.Conv2D(in_channels=channels, channels=nclass,
                                         kernel_size=1))

    def hybrid_forward(self, F, x):
        return self.block(x)


class _ConvBNReLU(HybridBlock):
    """Conv-BN-ReLU"""

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding=0,
                 norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(_ConvBNReLU, self).__init__()
        with self.name_scope():
            self.block = nn.HybridSequential()
            self.block.add(nn.Conv2D(in_channels=in_channels, channels=out_channels,
                                     kernel_size=kernel_size,
                                     padding=padding, strides=stride, use_bias=False))
            self.block.add(norm_layer(in_channels=out_channels,
                                      **({} if norm_kwargs is None else norm_kwargs)))
            self.block.add(nn.Activation('relu'))

    def hybrid_forward(self, F, x):
        return self.block(x)


class _DSConv(HybridBlock):
    """Depthwise Separable Convolutions"""
    def __init__(self, dw_channels, out_channels, stride=1,
                 norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(_DSConv, self).__init__()
        with self.name_scope():
            self.conv = nn.HybridSequential()
            self.conv.add(nn.Conv2D(in_channels=dw_channels, channels=dw_channels,
                                    kernel_size=3, strides=stride,
                                    padding=1, groups=dw_channels, use_bias=False))
            self.conv.add(norm_layer(in_channels=dw_channels,
                                     **({} if norm_kwargs is None else norm_kwargs)))
            self.conv.add(nn.Activation('relu'))
            self.conv.add(nn.Conv2D(in_channels=dw_channels, channels=out_channels,
                                    kernel_size=1, use_bias=False))
            self.conv.add(norm_layer(in_channels=out_channels,
                                     **({} if norm_kwargs is None else norm_kwargs)))
            self.conv.add(nn.Activation('relu'))

    def hybrid_forward(self, F, x):
        return self.conv(x)


class LearningToDownsample(HybridBlock):
    """Learning to downsample module"""
    def __init__(self, dw_channels1=32, dw_channels2=48, out_channels=64,
                 norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(LearningToDownsample, self).__init__()
        with self.name_scope():
            self.conv = _ConvBNReLU(3, dw_channels1, 3, 2,
                                    norm_layer=norm_layer, norm_kwargs=norm_kwargs, **kwargs)
            self.dsconv1 = _DSConv(dw_channels1, dw_channels2, 2,
                                   norm_layer=norm_layer, norm_kwargs=norm_kwargs, **kwargs)
            self.dsconv2 = _DSConv(dw_channels2, out_channels, 2,
                                   norm_layer=norm_layer, norm_kwargs=norm_kwargs, **kwargs)

    def hybrid_forward(self, F, x):
        x = self.conv(x)
        x = self.dsconv1(x)
        x = self.dsconv2(x)
        return x


class GlobalFeatureExtractor(HybridBlock):
    """Global feature extractor module"""
    def __init__(self, in_channels=64, block_channels=(64, 96, 128), out_channels=128, t=6,
                 num_blocks=(3, 3, 3), height=32, width=64,
                 norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(GlobalFeatureExtractor, self).__init__()
        self.num_blocks = num_blocks
        with self.name_scope():
            self.bottleneck1 = self._make_layer(LinearBottleneck, in_channels, block_channels[0],
                                                num_blocks[0], t, 2, norm_layer, norm_kwargs)
            self.bottleneck2 = self._make_layer(LinearBottleneck, block_channels[0],
                                                block_channels[1], num_blocks[1], t, 2,
                                                norm_layer, norm_kwargs)
            self.bottleneck3 = self._make_layer(LinearBottleneck, block_channels[1],
                                                block_channels[2], num_blocks[2], t, 1,
                                                norm_layer, norm_kwargs)
            self.ppm = _FastPyramidPooling(block_channels[2], out_channels,
                                           height=height, width=width,
                                           norm_layer=norm_layer, norm_kwargs=norm_kwargs)

    def _make_layer(self, block, inplanes, planes, blocks, t=6, stride=1,
                    norm_layer=nn.BatchNorm, norm_kwargs=None):
        layers = nn.HybridSequential()
        with layers.name_scope():
            layers.add(block(inplanes, planes, t, stride, norm_layer, norm_kwargs))
            for i in range(1, blocks):
                layers.add(block(planes, planes, t, 1, norm_layer, norm_kwargs))
        return layers

    def hybrid_forward(self, F, x):
        x = self.bottleneck1(x)
        x = self.bottleneck2(x)
        x = self.bottleneck3(x)
        x = self.ppm(x)
        return x


class LinearBottleneck(HybridBlock):
    """LinearBottleneck used in MobileNetV2"""
    def __init__(self, in_channels, out_channels, t=6, stride=2,
                 norm_layer=nn.BatchNorm, norm_kwargs=None):
        super(LinearBottleneck, self).__init__()
        self.use_shortcut = stride == 1 and in_channels == out_channels
        with self.name_scope():
            self.block = nn.HybridSequential()
            self.block.add(_ConvBNReLU(in_channels, in_channels * t, 1,
                                       norm_layer=norm_layer, norm_kwargs=norm_kwargs))
            self.block.add(_DWConv(in_channels * t, in_channels * t, stride,
                                   norm_layer=norm_layer, norm_kwargs=norm_kwargs))
            self.block.add(nn.Conv2D(in_channels=in_channels * t, channels=out_channels,
                                     kernel_size=1, use_bias=False))
            self.block.add(norm_layer(in_channels=out_channels,
                                      **({} if norm_kwargs is None else norm_kwargs)))

    def hybrid_forward(self, F, x):
        out = self.block(x)
        if self.use_shortcut:
            out = x + out
        return out


class _DWConv(HybridBlock):
    def __init__(self, dw_channels, out_channels, stride=1,
                 norm_layer=nn.BatchNorm, norm_kwargs=None):
        super(_DWConv, self).__init__()
        with self.name_scope():
            self.conv = nn.HybridSequential()
            self.conv.add(nn.Conv2D(in_channels=dw_channels, channels=out_channels, kernel_size=3,
                                    strides=stride, padding=1, groups=dw_channels, use_bias=False))
            self.conv.add(norm_layer(in_channels=out_channels,
                                     **({} if norm_kwargs is None else norm_kwargs)))
            self.conv.add(nn.Activation('relu'))

    def hybrid_forward(self, F, x):
        return self.conv(x)


def _PSP1x1Conv(in_channels, out_channels, norm_layer, norm_kwargs):
    block = nn.HybridSequential()
    with block.name_scope():
        block.add(nn.Conv2D(in_channels=in_channels, channels=out_channels,
                            kernel_size=1, use_bias=False))
        block.add(norm_layer(in_channels=out_channels,
                             **({} if norm_kwargs is None else norm_kwargs)))
        block.add(nn.Activation('relu'))
    return block


class _FastPyramidPooling(HybridBlock):
    def __init__(self, in_channels, ppm_out_channels, height=32, width=64,
                 norm_layer=nn.BatchNorm, norm_kwargs=None):
        super(_FastPyramidPooling, self).__init__()
        out_channels = int(in_channels/4)
        self._up_kwargs = {'height': height, 'width': width}
        with self.name_scope():
            self.conv1 = _PSP1x1Conv(in_channels, out_channels, norm_layer, norm_kwargs)
            self.conv2 = _PSP1x1Conv(in_channels, out_channels, norm_layer, norm_kwargs)
            self.conv3 = _PSP1x1Conv(in_channels, out_channels, norm_layer, norm_kwargs)
            self.conv4 = _PSP1x1Conv(in_channels, out_channels, norm_layer, norm_kwargs)
            self.out = _ConvBNReLU(in_channels * 2, ppm_out_channels, 1,
                                   norm_layer=norm_layer, norm_kwargs=norm_kwargs)

    def pool(self, F, x, size):
        return F.contrib.AdaptiveAvgPooling2D(x, output_size=size)

    def upsample(self, F, x):
        return F.contrib.BilinearResize2D(x, **self._up_kwargs)

    def hybrid_forward(self, F, x):
        feat1 = self.upsample(F, self.conv1(self.pool(F, x, 1)))
        feat2 = self.upsample(F, self.conv2(self.pool(F, x, 2)))
        feat3 = self.upsample(F, self.conv3(self.pool(F, x, 3)))
        feat4 = self.upsample(F, self.conv4(self.pool(F, x, 6)))
        x = F.concat(x, feat1, feat2, feat3, feat4, dim=1)
        x = self.out(x)
        return x

    def demo(self, x):
        """PyramidPooling for Fast SCNN """
        self._up_kwargs['height'] = x.shape[2]
        self._up_kwargs['width'] = x.shape[3]
        import mxnet.ndarray as F
        feat1 = self.upsample(F, self.conv1(self.pool(F, x, 1)))
        feat2 = self.upsample(F, self.conv2(self.pool(F, x, 2)))
        feat3 = self.upsample(F, self.conv3(self.pool(F, x, 3)))
        feat4 = self.upsample(F, self.conv4(self.pool(F, x, 6)))
        x = F.concat(x, feat1, feat2, feat3, feat4, dim=1)
        x = self.out(x)
        return x


def get_fastscnn(dataset='citys', ctx=cpu(0), pretrained=False,
                 root='~/.mxnet/models', **kwargs):
    r"""Fast-SCNN: Fast Semantic Segmentation Network
    Parameters
    ----------
    dataset : str, default cityscapes
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = get_fastscnn(dataset='citys')
    >>> print(model)
    """
    acronyms = {
        'citys': 'citys',
    }
    from ..data import datasets

    model = FastSCNN(datasets[dataset].NUM_CLASS, ctx=ctx, **kwargs)
    model.classes = datasets[dataset].classes

    if pretrained:
        from .model_store import get_model_file
        model.load_parameters(get_model_file('fastscnn_%s' % (acronyms[dataset]),
                                             tag=pretrained, root=root), ctx=ctx)
    return model


def get_fastscnn_citys(**kwargs):
    r"""Fast-SCNN: Fast Semantic Segmentation Network
        Parameters
        ----------
        dataset : str, default cityscapes
        ctx : Context, default CPU
            The context in which to load the pretrained weights.

        Examples
        --------
        >>> model = get_fastscnn_citys()
        >>> print(model)
        """
    return get_fastscnn('citys', **kwargs)



"""Utils for Semantic Segmentation"""
# pylint: disable=consider-using-enumerate,redefined-builtin,broad-except
import threading

from mxnet import autograd
from mxnet.gluon.utils import split_and_load
from mxnet.ndarray import NDArray

try:
    import Queue as queue
except ImportError:
    import queue

__all__ = ['DataParallelModel', 'DataParallelCriterion', 'parallel_backward', 'Parallelizable',
           'Parallel']


class Parallelizable(object):
    """Base class for parallelizable unit of work, which can be invoked by `Parallel`.
    The subclass must implement the `forward_backward` method, and be used
    together with `Parallel`. For example::
        class ParallelNet(Parallelizable):
            def __init__(self):
                self._net = Model()
                self._loss = gluon.loss.SoftmaxCrossEntropyLoss()
            def forward_backward(self, x):
                data, label = x
                with mx.autograd.record():
                    out = self._net(data)
                    loss = self._loss(out, label)
                loss.backward()
                return loss
        net = ParallelNet()
        ctx = [mx.gpu(0), mx.gpu(1)]
        parallel = Parallel(len(ctx), net)
        # Gluon block is initialized after forwarding the first batch
        initialized = False
        for batch in batches:
            for x in gluon.utils.split_and_load(batch, ctx):
                parallel.put(x)
            losses = [parallel.get() for _ in ctx]
            trainer.step()
    """

    def forward_backward(self, x):
        """ Forward and backward computation. """
        raise NotImplementedError()


class Parallel(object):
    """Class for parallel processing with `Parallelizable`s. It invokes a
    `Parallelizable` with multiple Python threads. For example::
        class ParallelNet(Parallelizable):
            def __init__(self):
                self._net = Model()
                self._loss = gluon.loss.SoftmaxCrossEntropyLoss()
            def forward_backward(self, x):
                data, label = x
                mx.autograd.record():
                    out = self._net(data)
                    loss = self._loss(out, label)
                loss.backward()
                return loss
        net = ParallelNet()
        ctx = [mx.gpu(0), mx.gpu(1)]
        parallel = Parallel(len(ctx), net)
        for batch in batches:
            for x in gluon.utils.split_and_load(batch, ctx):
                parallel.put(x)
            losses = [parallel.get() for _ in ctx]
            trainer.step()
    Parameters
    ----------
    num_workers : int
        Number of worker threads. If set to 0, the main thread is used as the worker for
        debugging purpose.
    parallelizable :
        Parallelizable net whose `forward` and `backward` methods are invoked
        by multiple worker threads.
    serial_init : bool, default True
        Execute the first `num_workers` inputs in main thread, so that the `Block`
        used in `parallizable` is initialized serially. Initialize a `Block` with
        multiple threads may cause unexpected behavior.
    """

    class _StopSignal:
        """Internal class to signal stop. """

        def __init__(self, msg):
            self._msg = msg

    def __init__(self, num_workers, parallizable, serial_init=True):
        self._in_queue = queue.Queue(-1)
        self._out_queue = queue.Queue(-1)
        self._num_workers = num_workers
        self._threads = []
        self._parallizable = parallizable
        self._num_serial = num_workers if serial_init else 0

        def _worker(in_queue, out_queue, parallel):
            while True:
                x = in_queue.get()
                if isinstance(x, Parallel._StopSignal):
                    return
                out = parallel.forward_backward(x)
                out_queue.put(out)

        arg = (self._in_queue, self._out_queue, self._parallizable)
        for _ in range(num_workers):
            thread = threading.Thread(target=_worker, args=arg)
            self._threads.append(thread)
            thread.start()

    def put(self, x):
        """Assign input `x` to an available worker and invoke
        `parallizable.forward_backward` with x. """
        if self._num_serial > 0 or len(self._threads) == 0:
            self._num_serial -= 1
            out = self._parallizable.forward_backward(x)
            self._out_queue.put(out)
        else:
            self._in_queue.put(x)

    def get(self):
        """Get an output of previous `parallizable.forward_backward` calls.
        This method blocks if none of previous `parallizable.forward_backward`
        calls have return any result. """
        return self._out_queue.get()

    def __del__(self):
        for thread in self._threads:
            if thread.is_alive():
                self._in_queue.put(self._StopSignal('stop'))
        for thread in self._threads:
            thread.join(10)


class DataParallelModel(object):
    """Data parallelism

    Hide the difference of single/multiple GPUs to the user.
    Inputs and outputs are both list of NDArrays in different contexts.
    In the forward pass, the module is replicated on each device,
    and each replica handles a portion of the input. During the backwards
    pass, gradients from each replica are summed into the original module.

    Parameters
    ----------
    module : object
        Network to be parallelized.
    ctx_list : list
        A list of contexts
    sync : bool
        enable synchronization (default: False).


    Inputs:
        - **inputs**: list of input (NDArrays)

    Outputs:
        - **outputs**: list of output (NDArrays)

    Example::
        >>> ctx = [mx.gpu(0), mx.gpu(1)]
        >>> net = DataParallelModel(model, ctx_list=ctx)
        >>> y = net(x)
    """

    def __init__(self, module, ctx_list=None, sync=False):
        module.collect_params().reset_ctx(ctx=ctx_list)
        self.ctx_list = ctx_list
        self.module = module
        self.sync = sync

    def __call__(self, *inputs, **kwargs):
        if not self.ctx_list:
            return self.module(*inputs, **kwargs)
        inputs, kwargs = split_load_kwargs(inputs, kwargs, self.ctx_list)
        assert (len(inputs) == len(self.ctx_list))
        if len(self.ctx_list) == 1:
            return tuple([tuple_map(self.module(*inputs[0], **kwargs[0]))])
        return parallel_apply(self.module, inputs, kwargs, self.sync)

    def __repr__(self):
        return 'DataParallel:\n module = {' + self.module.__repr__() + '}'


class DataParallelCriterion(object):
    """Criterion data parallelism

    Parameters
    ----------
    module : object
        Network to be parallelized.
    ctx : list
        A list of contexts to use.


    Inputs:

        - **inputs**: list of inputs (NDArrays)
        - **targets**: list of labels (NDArrays)

    Outputs:

        - **outputs**: list of output (NDArrays)

    Example::

        >>> ctx = [mx.gpu(0), mx.gpu(1)]
        >>> net = DataParallelModel(model, ctx=ctx)
        >>> criterion = DataParallelCriterion(criterion)
        >>> y = net(x)
        >>> losses = criterion(y, t)
    """

    def __init__(self, module, ctx_list=None, sync=False):
        self.module = module
        self.ctx_list = ctx_list
        self.sync = sync

    def __call__(self, inputs, *targets, **kwargs):
        # the inputs should be the outputs of DataParallelModel
        if not self.ctx_list:
            return self.module(inputs, *targets, **kwargs)
        targets, kwargs = split_load_kwargs(targets, kwargs, self.ctx_list)
        assert (len(targets) == len(self.ctx_list))
        if len(self.ctx_list) == 1:
            return tuple_map(self.module(*(inputs[0] + targets[0]), **kwargs[0]))
        assert (len(inputs) == len(self.ctx_list))
        return criterion_parallel_apply(self.module, inputs, targets, kwargs, self.sync)


def split_load_kwargs(inputs, kwargs, ctx_list, batch_axis=0):
    r"""Split with support for kwargs dictionary"""

    def split_map(obj):
        if isinstance(obj, NDArray):
            return split_and_load(obj, ctx_list, batch_axis, even_split=False)
        if isinstance(obj, tuple) and len(obj) > 0:
            return list(zip(*map(split_map, obj)))
        if isinstance(obj, list) and len(obj) > 0:
            return list(map(list, zip(*map(split_map, obj))))
        if isinstance(obj, dict) and len(obj) > 0:
            return list(map(type(obj), zip(*map(split_map, obj.items()))))
        return [obj for _ in ctx_list]

    inputs = split_map(inputs) if inputs else []
    kwargs = split_map(kwargs) if kwargs else []
    if len(inputs) < len(kwargs):
        inputs.extend([() for _ in range(len(kwargs) - len(inputs))])
    elif len(kwargs) < len(inputs):
        kwargs.extend([{} for _ in range(len(inputs) - len(kwargs))])
    inputs = tuple(inputs)
    kwargs = tuple(kwargs)
    return inputs, kwargs


def tuple_map(obj):
    if isinstance(obj, NDArray):
        return (obj,)
    if isinstance(obj, list) and len(obj) > 0:
        return tuple(obj)
    return obj


def parallel_apply(module, inputs, kwargs_tup=None, sync=False):
    """Parallel applying model forward"""
    if kwargs_tup is not None:
        assert len(inputs) == len(kwargs_tup)
    else:
        kwargs_tup = ({},) * len(inputs)

    lock = threading.Lock()
    results = {}

    def _worker(i, module, input, kwargs, results, is_recording, is_training, lock):
        try:
            if is_recording:
                with autograd.record(is_training):
                    output = tuple_map(module(*input, **kwargs))
                    for out in output:
                        out.wait_to_read()
            else:
                output = tuple_map(module(*input, **kwargs))
                for out in output:
                    out.wait_to_read()
            with lock:
                results[i] = output
        except Exception as e:
            with lock:
                results[i] = e

    is_training = autograd.is_training()
    is_recording = autograd.is_recording()
    threads = [threading.Thread(target=_worker,
                                args=(i, module, input, kwargs, results,
                                      is_recording, is_training, lock),)
               for i, (input, kwargs) in
               enumerate(zip(inputs, kwargs_tup))]

    if sync:
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        outputs = []
        for i in range(len(inputs)):
            output = results[i]
            if isinstance(output, Exception):
                raise output
            outputs.append(output)
        return tuple(outputs)
    else:
        outputs = [tuple_map(module(*input, **kwargs))
                   for (input, kwargs) in zip(inputs, kwargs_tup)]
        return tuple(outputs)


def criterion_parallel_apply(module, inputs, targets, kwargs_tup=None, sync=False):
    """Data Parallel Criterion"""
    if kwargs_tup:
        assert len(inputs) == len(kwargs_tup)
    else:
        kwargs_tup = ({},) * len(inputs)

    lock = threading.Lock()
    results = {}

    def _worker(i, module, input, target, kwargs, results, is_recording, is_training, lock):
        try:
            if is_recording:
                with autograd.record(is_training):
                    output = module(*(input + target), **kwargs)
                    output.wait_to_read()
            else:
                output = module(*(input + target), **kwargs)
                output.wait_to_read()
            with lock:
                results[i] = output
        except Exception as e:
            with lock:
                results[i] = e

    is_training = bool(autograd.is_training())
    is_recording = autograd.is_recording()

    threads = [threading.Thread(target=_worker,
                                args=(i, module, input, target,
                                      kwargs, results, is_recording, is_training, lock),)
               for i, (input, target, kwargs) in
               enumerate(zip(inputs, targets, kwargs_tup))]

    if sync:
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        outputs = []
        for i in range(len(inputs)):
            output = results[i]
            if isinstance(output, Exception):
                raise output
            outputs.append(output)
        return tuple(outputs)
    else:
        outputs = [module(*(input + target), **kwargs) \
                   for (input, target, kwargs) in zip(inputs, targets, kwargs_tup)]
        return tuple(outputs)


def parallel_backward(losses, sync=True):
    """Parallel Backward for CustomOp"""

    def _worker(loss):
        autograd.backward(loss)

    threads = [threading.Thread(target=_worker, args=(loss,)) for loss in losses]
    if sync:
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
    else:
        for loss in losses:
            loss.backward()



"""ResNet with Deconvolution layers for CenterNet object detection."""
# pylint: disable=unused-argument
from __future__ import absolute_import

import warnings
import math

import mxnet as mx
from mxnet.context import cpu
from mxnet.gluon import nn
from mxnet.gluon import contrib
from .. model_zoo import get_model

__all__ = ['DeconvResnet', 'get_deconv_resnet',
           'resnet18_v1b_deconv', 'resnet18_v1b_deconv_dcnv2',
           'resnet50_v1b_deconv', 'resnet50_v1b_deconv_dcnv2',
           'resnet101_v1b_deconv', 'resnet101_v1b_deconv_dcnv2']


class BilinearUpSample(mx.init.Initializer):
    """Initializes weights as bilinear upsampling kernel.

    Example
    -------
    >>> # Given 'module', an instance of 'mxnet.module.Module',
    initialize weights to bilinear upsample...
    >>> init = mx.initializer.BilinearUpSample()
    >>> module.init_params(init)
    >>> for dictionary in module.get_params():
    ...     for key in dictionary:
    ...         print(key)
    ...         print(dictionary[key].asnumpy())
    ...
    fullyconnected0_weight
    [[ 0.  0.  0.]]
    """
    def __init__(self):
        super(BilinearUpSample, self).__init__()

    def _init_weight(self, _, arr):
        mx.nd.random.normal(0, 0.01, arr.shape, out=arr)
        f = math.ceil(arr.shape[2] / 2)
        c = (2 * f - 1 - f % 2) / (2. * f)
        for i in range(arr.shape[2]):
            for j in range(arr.shape[3]):
                arr[0, 0, i, j] = \
                    (1 - math.fabs(i / f - c)) * (1 - math.fabs(j / f - c))
        for c in range(1, arr.shape[0]):
            arr[c, 0, :, :] = arr[0, 0, :, :]

class DeconvResnet(nn.HybridBlock):
    """Deconvolutional ResNet.

    Parameters
    ----------
    base_network : str
        Name of the base feature extraction network.
    deconv_filters : list of int
        Number of filters for deconv layers.
    deconv_kernels : list of int
        Kernel sizes for deconv layers.
    pretrained_base : bool
        Whether load pretrained base network.
    norm_layer : mxnet.gluon.nn.HybridBlock
        Type of Norm layers, can be BatchNorm, SyncBatchNorm, GroupNorm, etc.
    norm_kwargs : dict
        Additional kwargs for `norm_layer`.
    use_dcnv2 : bool
        If true, will use DCNv2 layers in upsampling blocks

    """
    def __init__(self, base_network='resnet18_v1b',
                 deconv_filters=(256, 128, 64), deconv_kernels=(4, 4, 4),
                 pretrained_base=True, norm_layer=nn.BatchNorm, norm_kwargs=None,
                 use_dcnv2=False, **kwargs):
        super(DeconvResnet, self).__init__(**kwargs)
        assert 'resnet' in base_network
        net = get_model(base_network, pretrained=pretrained_base)
        self._norm_layer = norm_layer
        self._norm_kwargs = norm_kwargs if norm_kwargs is not None else {}
        self._use_dcnv2 = use_dcnv2
        if 'v1b' in base_network:
            feat = nn.HybridSequential()
            feat.add(*[net.conv1,
                       net.bn1,
                       net.relu,
                       net.maxpool,
                       net.layer1,
                       net.layer2,
                       net.layer3,
                       net.layer4])
            self.base_network = feat
        else:
            raise NotImplementedError('Only v1 variants of resnet are supported so far.')
        with self.name_scope():
            self.deconv = self._make_deconv_layer(deconv_filters, deconv_kernels)

    def _get_deconv_cfg(self, deconv_kernel):
        """Get the deconv configs using presets"""
        if deconv_kernel == 4:
            padding = 1
            output_padding = 0
        elif deconv_kernel == 3:
            padding = 1
            output_padding = 1
        elif deconv_kernel == 2:
            padding = 0
            output_padding = 0
        else:
            raise ValueError('Unsupported deconvolution kernel: {}'.format(deconv_kernel))

        return deconv_kernel, padding, output_padding

    def _make_deconv_layer(self, num_filters, num_kernels):
        # pylint: disable=unused-variable
        """Make deconv layers using the configs"""
        assert len(num_kernels) == len(num_filters), \
            'Deconv filters and kernels number mismatch: {} vs. {}'.format(
                len(num_filters), len(num_kernels))

        layers = nn.HybridSequential('deconv_')
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.base_network.initialize()
        in_planes = self.base_network(mx.nd.zeros((1, 3, 256, 256))).shape[1]
        for planes, k in zip(num_filters, num_kernels):
            kernel, padding, output_padding = self._get_deconv_cfg(k)
            if self._use_dcnv2:
                assert hasattr(contrib.cnn, 'ModulatedDeformableConvolution'), \
                    "No ModulatedDeformableConvolution found in mxnet, consider upgrade..."
                layers.add(contrib.cnn.ModulatedDeformableConvolution(planes,
                                                                      kernel_size=3,
                                                                      strides=1,
                                                                      padding=1,
                                                                      dilation=1,
                                                                      num_deformable_group=1,
                                                                      in_channels=in_planes))
            else:
                layers.add(nn.Conv2D(channels=planes,
                                     kernel_size=3,
                                     strides=1,
                                     padding=1,
                                     in_channels=in_planes))
            layers.add(self._norm_layer(momentum=0.9, **self._norm_kwargs))
            layers.add(nn.Activation('relu'))
            layers.add(nn.Conv2DTranspose(channels=planes,
                                          kernel_size=kernel,
                                          strides=2,
                                          padding=padding,
                                          output_padding=output_padding,
                                          use_bias=False,
                                          in_channels=planes,
                                          weight_initializer=BilinearUpSample()))
            layers.add(self._norm_layer(momentum=0.9, **self._norm_kwargs))
            layers.add(nn.Activation('relu'))
            in_planes = planes

        return layers

    def hybrid_forward(self, F, x):
        # pylint: disable=arguments-differ
        """HybridForward"""
        y = self.base_network(x)
        out = self.deconv(y)
        return out

def get_deconv_resnet(base_network, pretrained=False, ctx=cpu(), use_dcnv2=False, **kwargs):
    """Get resnet with deconv layers.

    Parameters
    ----------
    base_network : str
        Name of the base feature extraction network.
    pretrained : bool
        Whether load pretrained base network.
    ctx : mxnet.Context
        mx.cpu() or mx.gpu()
    use_dcnv2 : bool
        If true, will use DCNv2 layers in upsampling blocks
    pretrained : type
        Description of parameter `pretrained`.
    Returns
    -------
    nn.HybridBlock
        Network instance of resnet with deconv layers

    """
    net = DeconvResnet(base_network=base_network, pretrained_base=pretrained,
                       use_dcnv2=use_dcnv2, **kwargs)
    with warnings.catch_warnings(record=True) as _:
        warnings.simplefilter("always")
        net.initialize()
    net.collect_params().reset_ctx(ctx)
    return net

def resnet18_v1b_deconv(**kwargs):
    """Resnet18 v1b model with deconv layers.

    Returns
    -------
    HybridBlock
        A Resnet18 v1b model with deconv layers.

    """
    kwargs['use_dcnv2'] = False
    return get_deconv_resnet('resnet18_v1b', **kwargs)

def resnet18_v1b_deconv_dcnv2(**kwargs):
    """Resnet18 v1b model with deconv layers and deformable v2 conv layers.

    Returns
    -------
    HybridBlock
        A Resnet18 v1b model with deconv layers and deformable v2 conv layers.

    """
    kwargs['use_dcnv2'] = True
    return get_deconv_resnet('resnet18_v1b', **kwargs)

def resnet50_v1b_deconv(**kwargs):
    """Resnet50 v1b model with deconv layers.

    Returns
    -------
    HybridBlock
        A Resnet50 v1b model with deconv layers.

    """
    kwargs['use_dcnv2'] = False
    return get_deconv_resnet('resnet50_v1b', **kwargs)

def resnet50_v1b_deconv_dcnv2(**kwargs):
    """Resnet50 v1b model with deconv layers and deformable v2 conv layers.

    Returns
    -------
    HybridBlock
        A Resnet50 v1b model with deconv layers and deformable v2 conv layers.

    """
    kwargs['use_dcnv2'] = True
    return get_deconv_resnet('resnet50_v1b', **kwargs)

def resnet101_v1b_deconv(**kwargs):
    """Resnet101 v1b model with deconv layers.

    Returns
    -------
    HybridBlock
        A Resnet101 v1b model with deconv layers.

    """
    kwargs['use_dcnv2'] = False
    return get_deconv_resnet('resnet101_v1b', **kwargs)

def resnet101_v1b_deconv_dcnv2(**kwargs):
    """Resnet101 v1b model with deconv layers and deformable v2 conv layers.

    Returns
    -------
    HybridBlock
        A Resnet101 v1b model with deconv layers and deformable v2 conv layers.

    """
    kwargs['use_dcnv2'] = True
    return get_deconv_resnet('resnet101_v1b', **kwargs)



# pylint: disable=missing-docstring,arguments-differ,line-too-long
"""Extended image transformations to video transformations.
Code adapted from:
https://github.com/bryanyzhu/two-stream-pytorch
https://github.com/facebookresearch/SlowFast
"""
from __future__ import division
import random
import math
import numbers
import numpy as np
from mxnet.gluon import Block

__all__ = ['VideoToTensor', 'VideoNormalize', 'VideoRandomHorizontalFlip', 'VideoMultiScaleCrop',
           'VideoCenterCrop', 'VideoTenCrop', 'VideoGroupTrainTransform', 'VideoGroupValTransform',
           'VideoGroupTrainTransformV2', 'VideoGroupValTransformV2', 'ShortSideRescale',
           'VideoGroupTrainTransformV3', 'RandomResizedCrop', 'VideoGroupTrainTransformV4']


class VideoGroupTrainTransform(Block):
    """Combination of transforms for training.
        (1) multiscale crop
        (2) scale
        (3) random horizontal flip
        (4) to tensor
        (5) normalize
    """
    def __init__(self, size, scale_ratios, mean, std, fix_crop=True,
                 more_fix_crop=True, max_distort=1, prob=0.5, max_intensity=255.0):
        super(VideoGroupTrainTransform, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = size[0]
        self.width = size[1]
        self.multiScaleCrop = VideoMultiScaleCrop(size=size,
                                                  scale_ratios=scale_ratios,
                                                  fix_crop=fix_crop,
                                                  more_fix_crop=more_fix_crop,
                                                  max_distort=max_distort)
        self.scale_ratios = scale_ratios
        self.fix_crop = fix_crop
        self.more_fix_crop = more_fix_crop
        self.max_distort = max_distort
        self.prob = prob
        self.max_intensity = max_intensity
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))


    def forward(self, clips):
        h, w, _ = clips[0].shape

        crop_size_pairs = self.multiScaleCrop.fillCropSize(self.scale_ratios, self.max_distort, h, w)
        size_sel = random.choice(crop_size_pairs)
        crop_height = size_sel[0]
        crop_width = size_sel[1]

        is_flip = random.random() < self.prob
        if self.fix_crop:
            offsets = self.multiScaleCrop.fillFixOffset(self.more_fix_crop, h, w, crop_height, crop_width)
            off_sel = random.choice(offsets)
            h_off = off_sel[0]
            w_off = off_sel[1]
        else:
            h_off = random.randint(0, h - crop_height)
            w_off = random.randint(0, w - crop_width)

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[h_off:h_off+crop_height, w_off:w_off+crop_width, :]
            scale_img = self.cv2.resize(crop_img, (self.width, self.height))
            if is_flip:
                flip_img = np.flip(scale_img, axis=1)
            else:
                flip_img = scale_img
            tensor_img = np.transpose(flip_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips


class VideoGroupValTransform(Block):
    """Combination of transforms for validation.
        (1) center crop
        (2) to tensor
        (3) normalize
    """

    def __init__(self, size, mean, std, max_intensity=255.0):
        super(VideoGroupValTransform, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))
        self.max_intensity = max_intensity

    def forward(self, clips):
        h, w, _ = clips[0].shape
        th, tw = self.size
        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[y1:y1+th, x1:x1+tw, :]
            tensor_img = np.transpose(crop_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips


class VideoToTensor(Block):
    """Convert images to tensor.

    Convert a list of images of shape (H x W x C) in the range
    [0, 255] to a float32 tensor of shape (C x H x W) in
    the range [0, 1).

    Parameters
    ----------
    max_intensity : float
        The maximum intensity value to be divided in order to fit the output tensor
        in the range [0, 1).

    Inputs:
        - **data**: a list of frames with shape [H x W x C] and uint8 type

    Outputs:
        - **out**: a list of frames with shape [C x H x W] and float32 type
    """
    def __init__(self, max_intensity=255.0):
        super(VideoToTensor, self).__init__()
        self.max_intensity = max_intensity

    def forward(self, clips):
        new_clips = []
        for cur_img in clips:
            new_clips.append(np.transpose(cur_img, axes=(2, 0, 1)) / self.max_intensity)
        return new_clips


class VideoNormalize(Block):
    """Normalize images with mean and standard deviation.

    Given mean `(m1, ..., mn)` and std `(s1, ..., sn)` for `n` channels,
    this transform normalizes each channel of the input tensor with::

        output[i] = (input[i] - mi) / si

    If mean or std is scalar, the same value will be applied to all channels.

    Parameters
    ----------
    mean : float or tuple of floats
        The mean values.
    std : float or tuple of floats
        The standard deviation values.

    Inputs:
        - **data**: a list of frames with shape [C x H x W]

    Outputs:
        - **out**: a list of normalized frames with shape [C x H x W]
    """

    def __init__(self, mean, std):
        super(VideoNormalize, self).__init__()
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))

    def forward(self, clips):
        new_clips = []
        for cur_img in clips:
            new_clips.append((cur_img - self.mean) / self.std)
        return new_clips


class VideoRandomHorizontalFlip(Block):
    """Randomly flip the images left to right with a probability.

    Parameters
    ----------
    prob : float
        The probability value to flip the images.

    Inputs:
        - **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of flipped frames with shape [H x W x C]
    """

    def __init__(self, prob=0.5):
        super(VideoRandomHorizontalFlip, self).__init__()
        self.prob = prob

    def forward(self, clips):
        new_clips = []
        if random.random() < self.prob:
            for cur_img in clips:
                new_clips.append(np.flip(cur_img, axis=1))
        else:
            new_clips = clips
        return new_clips


class ShortSideRescale(Block):
    """Short side rescale, keeping aspect ratio.

    Parameters
    ----------
    short_side : int
        A number that at least one side of the image has to be equal to.

    Inputs:
        - **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of rescaled frames with shape [short_side x (W*ratio) x C]
                   or [(H*ratio) x short_side x C]. Ratio is the input's original aspect ratio.
    """

    def __init__(self, short_side):
        super(ShortSideRescale, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.short_side = short_side

    def forward(self, clips):
        h, w, _ = clips[0].shape

        new_w = self.short_side
        new_h = self.short_side
        if w < h:
            new_h = int(math.floor((float(h) / w) * self.short_side))
        else:
            new_w = int(math.floor((float(w) / h) * self.short_side))

        new_clips = []
        for cur_img in clips:
            new_clips.append(self.cv2.resize(cur_img, (new_w, new_h)))
        return new_clips


class RandomResizedCrop(Block):
    """Random crop and resize.
    A crop of random size (default: of 0.08 to 1.0) of the original size and a random
    aspect ratio (default: of 3/4 to 4/3) of the original aspect ratio is made. This crop
    is finally resized to given size.
    This is popularly used to train the Inception networks.

    Parameters
    ----------
    size : tuple of int
        Expected output size of each edge
    scale : tuple of float
        Range of size of the origin size cropped
    ratio : tuple of float
        Range of aspect ratio of the origin aspect ratio cropped

    Inputs:
        - **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of cropped and resized frames with shape [size x size x C]
    """

    def __init__(self, size, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.)):
        super(RandomResizedCrop, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = size[0]
        self.width = size[1]
        self.scale = scale
        self.ratio = ratio

    def get_params(self, img, scale, ratio):
        """Get parameters for ``crop`` for a random sized crop.

        Parameters:
            img (cv2 Image): Image to be cropped.
            scale (tuple): range of size of the origin size cropped
            ratio (tuple): range of aspect ratio of the origin aspect ratio cropped
        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for a random
                sized crop.
        """
        height, width, _ = img.shape
        area = height * width

        for _ in range(10):
            target_area = random.uniform(*scale) * area
            aspect_ratio = random.uniform(*ratio)

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if random.random() < 0.5:
                w, h = h, w

            if 0 < w <= width and 0 < h <= height:
                i = random.randint(0, height - h)
                j = random.randint(0, width - w)
                return i, j, h, w

        # Fallback to center crop
        in_ratio = float(width) / float(height)
        if (in_ratio < min(ratio)):
            w = width
            h = int(round(w / min(ratio)))
        elif (in_ratio > max(ratio)):
            h = height
            w = int(round(h * max(ratio)))
        else:  # whole image
            w = width
            h = height

        i = (height - h) // 2
        j = (width - w) // 2
        return i, j, h, w

    def forward(self, clips):
        h_off, w_off, crop_height, crop_width = self.get_params(clips[0], self.scale, self.ratio)

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[h_off:h_off+crop_height, w_off:w_off+crop_width, :]
            new_clips.append(self.cv2.resize(crop_img, (self.width, self.height)))
        return new_clips


class VideoMultiScaleCrop(Block):
    """Corner cropping and multi-scale cropping.
    	Two data augmentation techniques introduced in:
        Towards Good Practices for Very Deep Two-Stream ConvNets,
        http://arxiv.org/abs/1507.02159
        Limin Wang, Yuanjun Xiong, Zhe Wang and Yu Qiao

    Parameters:
    ----------
    size : int
    	height and width required by network input, e.g., (224, 224)
    scale_ratios : list
    	efficient scale jittering, e.g., [1.0, 0.875, 0.75, 0.66]
    fix_crop : bool
    	use corner cropping or not. Default: True
    more_fix_crop : bool
    	use more corners or not. Default: True
    max_distort : float
    	maximum aspect ratio distortion, used together with scale_ratios. Default: 1

    Inputs:
    	- **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of cropped frames with shape [size x size x C]

    """

    def __init__(self, size, scale_ratios, fix_crop=True,
                 more_fix_crop=True, max_distort=1):
        super(VideoMultiScaleCrop, self).__init__()
        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = size[0]
        self.width = size[1]
        self.scale_ratios = scale_ratios
        self.fix_crop = fix_crop
        self.more_fix_crop = more_fix_crop
        self.max_distort = max_distort

    def fillFixOffset(self, more_fix_crop, image_h, image_w, crop_h, crop_w):
        """Fixed cropping strategy. Only crop the 4 corners and the center.
        If more_fix_crop is turned on, more corners will be counted.

        Inputs:
            - **data**: height and width of input tensor

        Outputs:
            - **out**: a list of locations to crop the image

        """
        h_off = (image_h - crop_h) // 4
        w_off = (image_w - crop_w) // 4

        offsets = []
        offsets.append((0, 0))          # upper left
        offsets.append((0, 4*w_off))    # upper right
        offsets.append((4*h_off, 0))    # lower left
        offsets.append((4*h_off, 4*w_off))  # lower right
        offsets.append((2*h_off, 2*w_off))  # center

        if more_fix_crop:
            offsets.append((0, 2*w_off))        # top center
            offsets.append((4*h_off, 2*w_off))  # bottom center
            offsets.append((2*h_off, 0))        # left center
            offsets.append((2*h_off, 4*w_off))  # right center

            offsets.append((1*h_off, 1*w_off))  # upper left quarter
            offsets.append((1*h_off, 3*w_off))  # upper right quarter
            offsets.append((3*h_off, 1*w_off))  # lower left quarter
            offsets.append((3*h_off, 3*w_off))  # lower right quarter

        return offsets

    def fillCropSize(self, scale_ratios, max_distort, image_h, image_w):
        """Fixed cropping strategy. Select crop size from
        pre-defined list (computed by scale_ratios).

        Inputs:
            - **data**: height and width of input tensor

        Outputs:
            - **out**: a list of crop sizes to crop the image

        """
        crop_sizes = []
        base_size = np.min((image_h, image_w))
        for h_index, scale_rate_h in enumerate(scale_ratios):
            crop_h = int(base_size * scale_rate_h)
            for w_index, scale_rate_w in enumerate(scale_ratios):
                crop_w = int(base_size * scale_rate_w)
                if (np.absolute(h_index - w_index) <= max_distort):
                    # To control the aspect ratio distortion
                    crop_sizes.append((crop_h, crop_w))

        return crop_sizes

    def forward(self, clips):
        h, w, _ = clips[0].shape

        crop_size_pairs = self.fillCropSize(self.scale_ratios, self.max_distort, h, w)
        size_sel = random.choice(crop_size_pairs)
        crop_height = size_sel[0]
        crop_width = size_sel[1]

        if self.fix_crop:
            offsets = self.fillFixOffset(self.more_fix_crop, h, w, crop_height, crop_width)
            off_sel = random.choice(offsets)
            h_off = off_sel[0]
            w_off = off_sel[1]
        else:
            h_off = random.randint(0, h - crop_height)
            w_off = random.randint(0, w - crop_width)

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[h_off:h_off+crop_height, w_off:w_off+crop_width, :]
            new_clips.append(self.cv2.resize(crop_img, (self.width, self.height)))
        return new_clips


class VideoCenterCrop(Block):
    """Crop images at the center to have a region of
    the given size. size can be a tuple (target_height, target_width)
    or an integer, in which case the target will be of a square shape (size, size)

    Parameters:
    ----------
    size : int
    	height and width required by network input, e.g., (224, 224)

    Inputs:
    	- **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of cropped frames with shape [size x size x C]
    """

    def __init__(self, size):
        super(VideoCenterCrop, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def forward(self, clips):
        h, w, _ = clips[0].shape
        th, tw = self.size
        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))

        new_clips = []
        for cur_img in clips:
            new_clips.append(cur_img[y1:y1+th, x1:x1+tw, :])
        return new_clips


class VideoThreeCrop(Block):
    """This method crops 3 regions. All regions will be in shape
    :obj`size`. Depending on the situation, these regions may consist of:
        (1) 1 center, 1 top and 1 bottom
        (2) 1 center, 1 left and 1 right

    Parameters:
    ----------
    size : int
        height and width required by network input, e.g., (224, 224)

    Inputs:
        - **data**: a list of N frames with shape [H x W x C]

    Outputs:
        - **out**: a list of 3xN cropped frames with shape [size x size x C]
    """

    def __init__(self, size):
        super(VideoThreeCrop, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def forward(self, clips):
        h, w, _ = clips[0].shape
        th, tw = self.size
        assert th == h or tw == w

        if th == h:
            w_step = (w - tw) // 2
            offsets = []
            offsets.append((0, 0))  # left
            offsets.append((2 * w_step, 0))  # right
            offsets.append((w_step, 0))  # middle
        elif tw == w:
            h_step = (h - th) // 2
            offsets = []
            offsets.append((0, 0))  # top
            offsets.append((0, 2 * h_step))  # down
            offsets.append((0, h_step))  # middle

        new_clips = []
        for ow, oh in offsets:
            for cur_img in clips:
                crop_img = cur_img[oh:oh+th, ow:ow+tw, :]
                new_clips.append(crop_img)
        return new_clips


class VideoTenCrop(Block):
    """Crop 10 regions from images.
    This is performed same as:
    http://chainercv.readthedocs.io/en/stable/reference/transforms.html#ten-crop

    This method crops 10 regions. All regions will be in shape
    :obj`size`. These regions consist of 1 center crop and 4 corner
    crops and horizontal flips of them.
    The crops are ordered in this order.
    * center crop
    * top-left crop
    * bottom-left crop
    * top-right crop
    * bottom-right crop
    * center crop (flipped horizontally)
    * top-left crop (flipped horizontally)
    * bottom-left crop (flipped horizontally)
    * top-right crop (flipped horizontally)
    * bottom-right crop (flipped horizontally)

    Parameters:
    ----------
    size : int
    	height and width required by network input, e.g., (224, 224)

    Inputs:
    	- **data**: a list of N frames with shape [H x W x C]

    Outputs:
        - **out**: a list of 10xN frames with shape [size x size x C]

    """
    def __init__(self, size):
        super(VideoTenCrop, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def forward(self, clips):
        h, w, _ = clips[0].shape
        oh, ow = self.size
        if h < oh or w < ow:
            raise ValueError("Cannot crop area {} from image with size \
            	({}, {})".format(str(self.size), h, w))

        new_clips = []
        for cur_img in clips:
            center = cur_img[(h - oh) // 2:(h + oh) // 2, (w - ow) // 2:(w + ow) // 2, :]
            tl = cur_img[0:oh, 0:ow, :]
            bl = cur_img[h - oh:h, 0:ow, :]
            tr = cur_img[0:oh, w - ow:w, :]
            br = cur_img[h - oh:h, w - ow:w, :]
            new_clips.append(center)
            new_clips.append(tl)
            new_clips.append(bl)
            new_clips.append(tr)
            new_clips.append(br)
            new_clips.append(np.flip(center, axis=1))
            new_clips.append(np.flip(tl, axis=1))
            new_clips.append(np.flip(bl, axis=1))
            new_clips.append(np.flip(tr, axis=1))
            new_clips.append(np.flip(br, axis=1))
        return new_clips


class VideoGroupTrainTransformV2(Block):
    """Combination of transforms for training.
    Follow the style of https://github.com/open-mmlab/mmaction
        (1) short side keep aspect ratio resize
        (2) multiscale crop
        (3) scale
        (4) random horizontal flip
        (5) to tensor
        (6) normalize
    """
    def __init__(self, size, short_side, scale_ratios, mean, std, fix_crop=True,
                 more_fix_crop=True, max_distort=1, prob=0.5, max_intensity=255.0):
        super(VideoGroupTrainTransformV2, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = size[0]
        self.width = size[1]
        self.short_side = short_side
        self.multiScaleCrop = VideoMultiScaleCrop(size=size,
                                                  scale_ratios=scale_ratios,
                                                  fix_crop=fix_crop,
                                                  more_fix_crop=more_fix_crop,
                                                  max_distort=max_distort)
        self.scale_ratios = scale_ratios
        self.fix_crop = fix_crop
        self.more_fix_crop = more_fix_crop
        self.max_distort = max_distort
        self.prob = prob
        self.max_intensity = max_intensity
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))

    def forward(self, clips):
        h, w, _ = clips[0].shape

        # step 1: short side keep aspect ratio resize
        new_w = self.short_side
        new_h = self.short_side
        if w < h:
            new_h = int(math.floor((float(h) / w) * self.short_side))
        else:
            new_w = int(math.floor((float(w) / h) * self.short_side))

        # step 2: multiscale crop
        crop_size_pairs = self.multiScaleCrop.fillCropSize(self.scale_ratios, self.max_distort, new_h, new_w)
        size_sel = random.choice(crop_size_pairs)
        crop_height = size_sel[0]
        crop_width = size_sel[1]

        is_flip = random.random() < self.prob
        if self.fix_crop:
            offsets = self.multiScaleCrop.fillFixOffset(self.more_fix_crop, new_h, new_w, crop_height, crop_width)
            off_sel = random.choice(offsets)
            h_off = off_sel[0]
            w_off = off_sel[1]
        else:
            h_off = random.randint(0, new_h - crop_height)
            w_off = random.randint(0, new_w - crop_width)

        new_clips = []
        for cur_img in clips:
            scale_img = self.cv2.resize(cur_img, (new_w, new_h))
            crop_img = scale_img[h_off:h_off+crop_height, w_off:w_off+crop_width, :]
            resize_img = self.cv2.resize(crop_img, (self.width, self.height))
            if is_flip:
                flip_img = np.flip(resize_img, axis=1)
            else:
                flip_img = resize_img
            tensor_img = np.transpose(flip_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips


class VideoGroupValTransformV2(Block):
    """Combination of transforms for validation.
    Follow the style of https://github.com/facebookresearch/SlowFast/
        (1) short side keep aspect ratio resize
        (2) center crop
        (3) to tensor
        (4) normalize
    """

    def __init__(self, crop_size, short_side, mean, std, max_intensity=255.0):
        super(VideoGroupValTransformV2, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = crop_size[0]
        self.width = crop_size[1]
        self.short_side = short_side
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))
        self.max_intensity = max_intensity

    def forward(self, clips):
        h, w, _ = clips[0].shape

        # step 1: short side keep aspect ratio resize
        new_w = self.short_side
        new_h = self.short_side
        if w < h:
            new_h = int(math.floor((float(h) / w) * self.short_side))
        else:
            new_w = int(math.floor((float(w) / h) * self.short_side))

        # step 2: center crop
        h_off = int(math.ceil((new_h - self.height) / 2))
        w_off = int(math.ceil((new_w - self.width) / 2))

        new_clips = []
        for cur_img in clips:
            scale_img = self.cv2.resize(cur_img, (new_w, new_h))
            crop_img = scale_img[h_off:h_off+self.height, w_off:w_off+self.width, :]
            tensor_img = np.transpose(crop_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips


class VideoGroupTrainTransformV3(Block):
    """Combination of transforms for training.
    Follow the style of https://github.com/facebookresearch/SlowFast/
        (1) random short side scale jittering
        (2) random crop
        (3) random horizontal flip
        (4) to tensor
        (5) normalize
    """
    def __init__(self, crop_size, min_size, max_size,
                 mean, std, prob=0.5, max_intensity=255.0):
        super(VideoGroupTrainTransformV3, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = crop_size[0]
        self.width = crop_size[1]
        self.min_size = min_size
        self.max_size = max_size
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))
        self.prob = prob
        self.max_intensity = max_intensity

    def forward(self, clips):
        h, w, _ = clips[0].shape

        # step 1: random short side scale jittering
        size = int(round(np.random.uniform(self.min_size, self.max_size)))
        new_w = size
        new_h = size
        if w < h:
            new_h = int(math.floor((float(h) / w) * size))
        else:
            new_w = int(math.floor((float(w) / h) * size))

        # step 2: random crop
        h_off = 0
        if new_h > self.height:
            h_off = int(np.random.randint(0, new_h - self.height))
        w_off = 0
        if new_w > self.width:
            w_off = int(np.random.randint(0, new_w - self.width))

        # step 3: random horizontal flip
        is_flip = random.random() < self.prob

        new_clips = []
        for cur_img in clips:
            scale_img = self.cv2.resize(cur_img, (new_w, new_h))
            crop_img = scale_img[h_off:h_off+self.height, w_off:w_off+self.width, :]
            if is_flip:
                flip_img = np.flip(crop_img, axis=1)
            else:
                flip_img = crop_img
            tensor_img = np.transpose(flip_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips


class VideoGroupTrainTransformV4(Block):
    """Combination of transforms for training.
    Follow the style of https://github.com/open-mmlab/mmaction.
    This is only for tranining SlowFast family networks in mmaction.
        (1) random crop and resize
        (2) random horizontal flip
        (3) to tensor
        (4) normalize
    """
    def __init__(self, size, mean, std, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.),
                 prob=0.5, max_intensity=255.0):
        super(VideoGroupTrainTransformV4, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = size[0]
        self.width = size[1]
        self.randomResizedCrop = RandomResizedCrop(size=size,
                                                   scale=scale,
                                                   ratio=ratio)
        self.scale = scale
        self.ratio = ratio
        self.prob = prob
        self.max_intensity = max_intensity
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))

    def forward(self, clips):
        h_off, w_off, crop_height, crop_width = self.randomResizedCrop.get_params(clips[0], self.scale, self.ratio)
        is_flip = random.random() < self.prob

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[h_off:h_off+crop_height, w_off:w_off+crop_width, :]
            resize_img = self.cv2.resize(crop_img, (self.width, self.height))
            if is_flip:
                flip_img = np.flip(resize_img, axis=1)
            else:
                flip_img = resize_img
            tensor_img = np.transpose(flip_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips



"""MSCOCO Semantic Segmentation pretraining for VOC."""
import os
import pickle
from tqdm import trange
from PIL import Image
import numpy as np

from .utils import try_import_pycocotools
from ..segbase import SegmentationDataset

class COCOSegmentation(SegmentationDataset):
    """COCO Semantic Segmentation Dataset for VOC Pre-training.

    Parameters
    ----------
    root : string
        Path to COCO folder. Default is '$(HOME)/mxnet/datasets/coco'
    split: string
        'train', 'val' or 'test'
    transform : callable, optional
        A function that transforms the image

    Examples
    --------
    >>> from mxnet.gluon.data.vision import transforms
    >>> # Transforms for Normalization
    >>> input_transform = transforms.Compose([
    >>>     transforms.ToTensor(),
    >>>     transforms.Normalize([.485, .456, .406], [.229, .224, .225]),
    >>> ])
    >>> # Create Dataset
    >>> trainset = gluoncv.data.COCOSegmentation(split='train', transform=input_transform)
    >>> # Create Training Loader
    >>> train_data = gluon.data.DataLoader(
    >>>     trainset, 4, shuffle=True, last_batch='rollover',
    >>>     num_workers=4)
    """
    CAT_LIST = [0, 5, 2, 16, 9, 44, 6, 3, 17, 62, 21, 67, 18, 19, 4,
                1, 64, 20, 63, 7, 72]
    NUM_CLASS = 21
    CLASSES = ("background", "airplane", "bicycle", "bird", "boat", "bottle",
               "bus", "car", "cat", "chair", "cow", "diningtable", "dog", "horse",
               "motorcycle", "person", "potted-plant", "sheep", "sofa", "train",
               "tv")

    def __init__(self, root=os.path.expanduser('~/.mxnet/datasets/coco'),
                 split='train', mode=None, transform=None, **kwargs):
        super(COCOSegmentation, self).__init__(root, split, mode, transform, **kwargs)
        # lazy import pycocotools
        try_import_pycocotools()
        from pycocotools.coco import COCO
        from pycocotools import mask
        if split == 'train':
            print('train set')
            ann_file = os.path.join(root, 'annotations/instances_train2017.json')
            ids_file = os.path.join(root, 'annotations/train_ids.mx')
            self.root = os.path.join(root, 'train2017')
        else:
            print('val set')
            ann_file = os.path.join(root, 'annotations/instances_val2017.json')
            ids_file = os.path.join(root, 'annotations/val_ids.mx')
            self.root = os.path.join(root, 'val2017')
        self.coco = COCO(ann_file)
        self.coco_mask = mask
        if os.path.exists(ids_file):
            with open(ids_file, 'rb') as f:
                self.ids = pickle.load(f)
        else:
            ids = list(self.coco.imgs.keys())
            self.ids = self._preprocess(ids, ids_file)
        self.transform = transform

    def __getitem__(self, index):
        coco = self.coco
        img_id = self.ids[index]
        img_metadata = coco.loadImgs(img_id)[0]
        path = img_metadata['file_name']
        img = Image.open(os.path.join(self.root, path)).convert('RGB')
        cocotarget = coco.loadAnns(coco.getAnnIds(imgIds=img_id))
        mask = Image.fromarray(self._gen_seg_mask(
            cocotarget, img_metadata['height'], img_metadata['width']))
        # synchrosized transform
        if self.mode == 'train':
            img, mask = self._sync_transform(img, mask)
        elif self.mode == 'val':
            img, mask = self._val_sync_transform(img, mask)
        else:
            assert self.mode == 'testval'
            img, mask = self._img_transform(img), self._mask_transform(mask)
        # general resize, normalize and toTensor
        if self.transform is not None:
            img = self.transform(img)
        return img, mask

    def __len__(self):
        return len(self.ids)

    def _gen_seg_mask(self, target, h, w):
        mask = np.zeros((h, w), dtype=np.uint8)
        coco_mask = self.coco_mask
        for instance in target:
            rle = coco_mask.frPyObjects(instance['segmentation'], h, w)
            m = coco_mask.decode(rle)
            cat = instance['category_id']
            if cat in self.CAT_LIST:
                c = self.CAT_LIST.index(cat)
            else:
                continue
            if len(m.shape) < 3:
                mask[:, :] += (mask == 0) * (m * c)
            else:
                mask[:, :] += (mask == 0) * (((np.sum(m, axis=2)) > 0) * c).astype(np.uint8)
        return mask

    def _preprocess(self, ids, ids_file):
        print("Preprocessing mask, this will take a while." + \
              "But don't worry, it only run once for each split.")
        tbar = trange(len(ids))
        new_ids = []
        for i in tbar:
            img_id = ids[i]
            cocotarget = self.coco.loadAnns(self.coco.getAnnIds(imgIds=img_id))
            img_metadata = self.coco.loadImgs(img_id)[0]
            mask = self._gen_seg_mask(cocotarget, img_metadata['height'],
                                      img_metadata['width'])
            # more than 1k pixels
            if (mask > 0).sum() > 1000:
                new_ids.append(img_id)
            tbar.set_description('Doing: {}/{}, got {} qualified images'.\
                format(i, len(ids), len(new_ids)))
        print('Found number of qualified images: ', len(new_ids))
        with open(ids_file, 'wb') as f:
            pickle.dump(new_ids, f)
        return new_ids

    @property
    def classes(self):
        """Category names."""
        return type(self).CLASSES



"""Utility functions for version checking."""
import sys
import warnings

__all__ = ['check_version', '_require_mxnet_version', '_deprecate_python2']

def check_version(min_version, warning_only=False):
    """Check the version of gluoncv satisfies the provided minimum version.
    An exception is thrown if the check does not pass.

    Parameters
    ----------
    min_version : str
        Minimum version
    warning_only : bool
        Printing a warning instead of throwing an exception.
    """
    from .. import __version__
    from distutils.version import LooseVersion
    bad_version = LooseVersion(__version__) < LooseVersion(min_version)
    if bad_version:
        msg = 'Installed GluonCV version (%s) does not satisfy the ' \
              'minimum required version (%s)'%(__version__, min_version)
        if warning_only:
            warnings.warn(msg)
        else:
            raise AssertionError(msg)


def _require_mxnet_version(mx_version, max_mx_version='2.0.0'):
    try:
        import mxnet as mx
        from distutils.version import LooseVersion
        if LooseVersion(mx.__version__) < LooseVersion(mx_version) or \
            LooseVersion(mx.__version__) >= LooseVersion(max_mx_version):
            version_str = '>={},<{}'.format(mx_version, max_mx_version)
            msg = (
                "Legacy mxnet-mkl=={0} detected, some modules may not work properly. "
                "mxnet-mkl{1} is required. You can use pip to upgrade mxnet "
                "`pip install -U 'mxnet-mkl{1}'` "
                "or `pip install -U 'mxnet-cu100mkl{1}'`\
                ").format(mx.__version__, version_str)
            raise RuntimeError(msg)
    except ImportError:
        raise ImportError(
            "Unable to import dependency mxnet. "
            "A quick tip is to install via "
            "`pip install 'mxnet-cu100mkl<{}'`. "
            "please refer to https://gluon-cv.mxnet.io/#installation for details.".format(
                max_mx_version))

def _deprecate_python2():
    if sys.version_info[0] < 3:
        msg = 'Python2 has reached the end of its life on January 1st, 2020. ' + \
            'A future version of gluoncv will drop support for Python 2.'
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(msg, DeprecationWarning)



import mxnet as mx
from mxnet.gluon import nn
from mxnet.gluon.nn import HybridBlock
# pylint: disable-all

__all__ = ['PAM_Module', 'CAM_Module']


class PAM_Module(HybridBlock):
    r""" Position attention module
    from the paper `"Dual Attention Network for Scene Segmentation"
    <https://arxiv.org/abs/1809.02983>`
    PAM_Module captures long-range spatial contextual information.
    """
    def __init__(self, in_dim):
        super(PAM_Module, self).__init__()
        self.chanel_in = in_dim

        self.query_conv = nn.Conv2D(in_channels=in_dim, channels=in_dim//8, kernel_size=(1, 1))
        self.key_conv = nn.Conv2D(in_channels=in_dim, channels=in_dim//8, kernel_size=(1, 1))
        self.value_conv = nn.Conv2D(in_channels=in_dim, channels=in_dim, kernel_size=(1, 1))
        self.gamma = self.params.get('gamma', shape=(1,), init=mx.init.Zero())

    def hybrid_forward(self, F, x, **kwargs):
        """
            inputs :
                x : input feature maps( B X C X H X W)
            returns :
                out : attention value + input feature
                attention: B X (HxW) X (HxW)
        """
        gamma = kwargs['gamma']
        proj_query = F.reshape(self.query_conv(x), (0, 0, -1))
        proj_key = F.reshape(self.key_conv(x), (0, 0, -1))
        energy = F.batch_dot(proj_query, proj_key, transpose_a=True)
        attention = F.softmax(energy)
        proj_value = F.reshape(self.value_conv(x), (0, 0, -1))
        out = F.batch_dot(proj_value, attention, transpose_b=True)
        out = F.reshape_like(out, x, lhs_begin=2, lhs_end=None, rhs_begin=2, rhs_end=None)

        out = F.broadcast_mul(gamma, out) + x

        return out


class CAM_Module(HybridBlock):
    r""" Channel attention module
    from the paper `"Dual Attention Network for Scene Segmentation"
    <https://arxiv.org/abs/1809.02983>`
    CAM_Module explicitly models interdependencies between channels.
    """
    def __init__(self, in_dim):
        super(CAM_Module, self).__init__()
        self.chanel_in = in_dim

        self.gamma = self.params.get('gamma', shape=(1,), init=mx.init.Zero())

    def hybrid_forward(self, F, x, **kwargs):
        """
            inputs :
                x : input feature maps( B X C X H X W)
            returns :
                out : attention value + input feature
                attention: B X C X C
        """
        gamma = kwargs['gamma']
        proj_query = F.reshape(x, (0, 0, -1))
        proj_key = F.reshape(x, (0, 0, -1))
        energy = F.batch_dot(proj_query, proj_key, transpose_b=True)
        energy_new = F.max(energy, -1, True).broadcast_like(energy) - energy
        attention = F.softmax(energy_new)
        proj_value = F.reshape(x, (0, 0, -1))

        out = F.batch_dot(attention, proj_value)
        out = F.reshape_like(out, x, lhs_begin=2, lhs_end=None, rhs_begin=2, rhs_end=None)

        out = F.broadcast_mul(gamma, out) + x
        return out



# pylint: disable=wildcard-import
"""Mask RCNN Instance Segmentation"""
from __future__ import absolute_import

from .mask_rcnn import *
from .predefined_models import *
from .data_parallel import ForwardBackwardTask



""" SiamRPN metrics """
import numpy as np
from ..filesystem import try_import_colorama

def Iou(rect1, rect2):
    """
    caculate interection over union

    Parameters
    ----------
        rect1: list or np.array, rectangle1
        rect2: list or np.array, rectangle2

    Returns
    -------
    iou
    """
    x1, y1, x2, y2 = rect1[0], rect1[1], rect1[2], rect1[3]
    tx1, ty1, tx2, ty2 = rect2[0], rect2[1], rect2[2], rect2[3]

    xx1 = np.maximum(tx1, x1)
    yy1 = np.maximum(ty1, y1)
    xx2 = np.minimum(tx2, x2)
    yy2 = np.minimum(ty2, y2)

    ww = np.maximum(0, xx2 - xx1)
    hh = np.maximum(0, yy2 - yy1)

    area = (x2-x1) * (y2-y1)
    target_a = (tx2-tx1) * (ty2 - ty1)
    inter = ww * hh
    iou = inter / (area + target_a - inter)
    return iou

def overlap_ratio(rect1, rect2):
    """Compute overlap ratio between two rects

    Parameters
    ----------
        rect1 : nd.array
            2d array of N x [x,y,w,h]
        rect2 : nd.array
            2d array of N x [x,y,w,h]

    Return
    ----------
        IOU
    """
    left = np.maximum(rect1[:, 0], rect2[:, 0])
    right = np.minimum(rect1[:, 0]+rect1[:, 2], rect2[:, 0]+rect2[:, 2])
    top = np.maximum(rect1[:, 1], rect2[:, 1])
    bottom = np.minimum(rect1[:, 1]+rect1[:, 3], rect2[:, 1]+rect2[:, 3])

    intersect = np.maximum(0, right - left) * np.maximum(0, bottom - top)
    union = rect1[:, 2]*rect1[:, 3] + rect2[:, 2]*rect2[:, 3] - intersect
    iou = intersect / union
    iou = np.maximum(np.minimum(1, iou), 0)
    return iou

def success_overlap(gt_bb, result_bb, n_frame):
    """get success_overlap score

    Parameters
    ----------
        result_bb : nd.array
            2d array of N x [x,y,w,h]
        n_frame : int
            frame number

    Return
    ----------
        success score
    """
    thresholds_overlap = np.arange(0, 1.05, 0.05)
    success = np.zeros(len(thresholds_overlap))
    iou = np.ones(len(gt_bb)) * (-1)
    mask = np.sum(gt_bb[:, 2:] > 0, axis=1) == 2
    iou[mask] = overlap_ratio(gt_bb[mask], result_bb[mask])
    for i, per_threshold in enumerate(thresholds_overlap):
        success[i] = np.sum(iou > per_threshold) / float(n_frame)
    return success

def success_error(gt_center, result_center, thresholds, n_frame):
    """get success_error score

    Parameters
    ----------
        gt_center : np.ndarray
            2d array of N x [x,y,w,h]
        result_center : np.ndarray
            2d array of N x [x,y,w,h]
        thresholds : float
            error float
        n_frame : int
            frame number

    Return
    ----------
        success_error score
    """
    success = np.zeros(len(thresholds))
    dist = np.ones(len(gt_center)) * (-1)
    mask = np.sum(gt_center > 0, axis=1) == 2
    dist[mask] = np.sqrt(np.sum(
        np.power(gt_center[mask] - result_center[mask], 2), axis=1))
    for i, per_threshold in enumerate(thresholds):
        success[i] = np.sum(dist <= per_threshold) / float(n_frame)
    return success


class OPEBenchmark:
    """
    SiamRPN OPEBenchmark have eval_success, precision to select.
    eval_success is  distance between the center point of the predicted position
    precision is Compute overlap ratio between two rects through thresholds_overlap
    Parameters
    ----------
        dataset
            dataset Benchmark
    """
    def __init__(self, dataset):
        self.dataset = dataset

    def convert_bb_to_center(self, bboxes):
        return np.array([(bboxes[:, 0] + (bboxes[:, 2] - 1) / 2),
                         (bboxes[:, 1] + (bboxes[:, 3] - 1) / 2)]).T

    def convert_bb_to_norm_center(self, bboxes, gt_wh):
        return self.convert_bb_to_center(bboxes) / (gt_wh+1e-16)

    def eval_success(self, eval_trackers=None):
        """eval_success is  distance between the center point of the predicted position
           and the center position marked in the benchmark
        Parameters
        ----------
            eval_trackers: list of tracker name or single tracker name
        Return
        ----------
            return: dict of results
        """
        if eval_trackers is None:
            eval_trackers = self.dataset.tracker_names
        if isinstance(eval_trackers, str):
            eval_trackers = [eval_trackers]

        success_ret = {}
        for tracker_name in eval_trackers:
            success_ret_ = {}
            for video in self.dataset:
                gt_traj = np.array(video.gt_traj)
                if tracker_name not in video.pred_trajs:
                    tracker_traj = video.load_tracker(self.dataset.tracker_path,
                                                      tracker_name, False)
                    tracker_traj = np.array(tracker_traj)
                else:
                    tracker_traj = np.array(video.pred_trajs[tracker_name])
                n_frame = len(gt_traj)
                if hasattr(video, 'absent'):
                    gt_traj = gt_traj[video.absent == 1]
                    tracker_traj = tracker_traj[video.absent == 1]
                success_ret_[video.name] = success_overlap(gt_traj, tracker_traj, n_frame)
            success_ret[tracker_name] = success_ret_
        return success_ret

    def eval_precision(self, eval_trackers=None):
        """
        eval model precision in eval_precision
        Parameters
        ----------
            eval_trackers: list of tracker name or single tracker name
        Return
        ----------
            return: dict of results
        """
        if eval_trackers is None:
            eval_trackers = self.dataset.tracker_names
        if isinstance(eval_trackers, str):
            eval_trackers = [eval_trackers]

        precision_ret = {}
        for tracker_name in eval_trackers:
            precision_ret_ = {}
            for video in self.dataset:
                gt_traj = np.array(video.gt_traj)
                if tracker_name not in video.pred_trajs:
                    tracker_traj = video.load_tracker(self.dataset.tracker_path,
                                                      tracker_name, False)
                    tracker_traj = np.array(tracker_traj)
                else:
                    tracker_traj = np.array(video.pred_trajs[tracker_name])
                n_frame = len(gt_traj)
                if hasattr(video, 'absent'):
                    gt_traj = gt_traj[video.absent == 1]
                    tracker_traj = tracker_traj[video.absent == 1]
                gt_center = self.convert_bb_to_center(gt_traj)
                tracker_center = self.convert_bb_to_center(tracker_traj)
                thresholds = np.arange(0, 51, 1)
                precision_ret_[video.name] = success_error(gt_center, tracker_center,
                                                           thresholds, n_frame)
            precision_ret[tracker_name] = precision_ret_
        return precision_ret

    def eval_norm_precision(self, eval_trackers=None):
        """
        eval model precision in eval_norm_precision
        Parameters
        ----------
            eval_trackers: list of tracker name or single tracker name
        Return
        ----------
            return: dict of results
        """
        if eval_trackers is None:
            eval_trackers = self.dataset.tracker_names
        if isinstance(eval_trackers, str):
            eval_trackers = [eval_trackers]

        norm_precision_ret = {}
        for tracker_name in eval_trackers:
            norm_precision_ret_ = {}
            for video in self.dataset:
                gt_traj = np.array(video.gt_traj)
                if tracker_name not in video.pred_trajs:
                    tracker_traj = video.load_tracker(self.dataset.tracker_path,
                                                      tracker_name, False)
                    tracker_traj = np.array(tracker_traj)
                else:
                    tracker_traj = np.array(video.pred_trajs[tracker_name])
                n_frame = len(gt_traj)
                if hasattr(video, 'absent'):
                    gt_traj = gt_traj[video.absent == 1]
                    tracker_traj = tracker_traj[video.absent == 1]
                gt_center_norm = self.convert_bb_to_norm_center(gt_traj, gt_traj[:, 2:4])
                tracker_center_norm = self.convert_bb_to_norm_center(tracker_traj, gt_traj[:, 2:4])
                thresholds = np.arange(0, 51, 1) / 100
                norm_precision_ret_[video.name] = success_error(gt_center_norm,
                                                                tracker_center_norm,
                                                                thresholds,
                                                                n_frame)
            norm_precision_ret[tracker_name] = norm_precision_ret_
        return norm_precision_ret

    def show_result(self, success_ret, precision_ret=None,
                    norm_precision_ret=None, show_video_level=False, helight_threshold=0.6):
        """pretty print result
        Parameters
        ----------
            success_ret: returned dict from function eval
        """
        colorama = try_import_colorama()
        tracker_auc = {}
        for tracker_name in success_ret.keys():
            auc = np.mean(list(success_ret[tracker_name].values()))
            tracker_auc[tracker_name] = auc
        tracker_auc_ = sorted(tracker_auc.items(),
                              key=lambda x: x[1],
                              reverse=True)[:20]
        tracker_names = [x[0] for x in tracker_auc_]


        tracker_name_len = max((max([len(x) for x in success_ret.keys()])+2), 12)
        header = ("|{:^"+str(tracker_name_len)+"}|{:^9}|{:^16}|{:^11}|").format("Tracker name",
                                                                                "Success",
                                                                                "Norm Precision",
                                                                                "Precision")
        formatter = "|{:^"+str(tracker_name_len)+"}|{:^9.3f}|{:^16.3f}|{:^11.3f}|"
        print('-'*len(header))
        print(header)
        print('-'*len(header))
        for tracker_name in tracker_names:
            success = tracker_auc[tracker_name]
            if precision_ret is not None:
                precision = np.mean(list(precision_ret[tracker_name].values()), axis=0)[20]
            else:
                precision = 0
            if norm_precision_ret is not None:
                norm_precision = np.mean(list(norm_precision_ret[tracker_name].values()),
                                         axis=0)[20]
            else:
                norm_precision = 0
            print(formatter.format(tracker_name, success, norm_precision, precision))
        print('-'*len(header))

        if show_video_level and len(success_ret) < 10 \
                and precision_ret is not None \
                and len(precision_ret) < 10:
            print("\n\n")
            header1 = "|{:^21}|".format("Tracker name")
            header2 = "|{:^21}|".format("Video name")
            for tracker_name in success_ret.keys():
                header1 += ("{:^21}|").format(tracker_name)
                header2 += "{:^9}|{:^11}|".format("success", "precision")
            print('-'*len(header1))
            print(header1)
            print('-'*len(header1))
            print(header2)
            print('-'*len(header1))
            videos = list(success_ret[tracker_name].keys())
            for video in videos:
                row = "|{:^21}|".format(video)
                for tracker_name in success_ret.keys():
                    success = np.mean(success_ret[tracker_name][video])
                    precision = np.mean(precision_ret[tracker_name][video])
                    success_str = "{:^9.3f}".format(success)
                    if success < helight_threshold:
                        row += f'{colorama.Fore.RED}{success_str}{colorama.Style.RESET_ALL}|'
                    else:
                        row += success_str+'|'
                    precision_str = "{:^11.3f}".format(precision)
                    if precision < helight_threshold:
                        row += f'{colorama.Fore.RED}{precision_str}{colorama.Style.RESET_ALL}|'
                    else:
                        row += precision_str+'|'
                print(row)
            print('-'*len(header1))
